package com.zanxiang.game.data.serve.service.impl;

import com.zanxiang.game.data.serve.pojo.entity.AdsOverallSummary;
import com.zanxiang.game.data.serve.pojo.param.AdsOverallSummaryParam;
import com.zanxiang.game.data.serve.pojo.vo.AdsOverallSummaryLineDataVO;
import com.zanxiang.game.data.serve.pojo.vo.AdsOverallSummaryVO;
import com.zanxiang.game.data.serve.service.IAdsOverallSummaryService;
import lombok.extern.slf4j.Slf4j;
import org.nutz.dao.Cnd;
import org.nutz.dao.Condition;
import org.nutz.dao.Dao;
import org.nutz.dao.Sqls;
import org.nutz.dao.sql.Criteria;
import org.nutz.dao.sql.Sql;
import org.springframework.beans.BeanUtils;
import org.springframework.beans.BeanWrapper;
import org.springframework.beans.BeanWrapperImpl;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Service;

import java.beans.PropertyDescriptor;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.math.RoundingMode;
import java.time.LocalDate;
import java.time.temporal.ChronoUnit;
import java.util.*;

/**
 * @author tianhua
 * @time 2023/7/4
 * @Description 整体概况逻辑处理
 **/
@Service
@Slf4j
public class AdsOverallSummaryServiceImpl implements IAdsOverallSummaryService {

    @Autowired
    private Dao dao;

    @Override
    public AdsOverallSummaryVO getOverallSummaryData(AdsOverallSummaryParam param) {

        //获取查询条件 整体数据查询 不需要日期
        Condition totalDataCon = myCondition(param, false, -1);
        //查询条件转换为sql语句 设置映射 Entity为 AdsOverallSummary对象
        String totalDataConString = totalDataCon.toSql(dao.getEntity(AdsOverallSummary.class));
        //自定义sql语句
        Sql totalDataSql = Sqls.create("SELECT \n" +
                "IFNULL(SUM(x.today_player_count),0) total_players ,\n" +
                "IFNULL(SUM(x.today_total_cost),0) total_cost ,\n" +
                "IFNULL(SUM(x.today_amount),0) total_amount ,\n" +
                "IFNULL(SUM(x.today_agent_count),0) total_agent_count ,\n" +
                "IFNULL(SUM(x.today_game_count),0) total_game_count ,\n" +
                "IFNULL(SUM(x.today_new_player_amount)/SUM(x.today_total_cost),0) total_roi \n" +
                "FROM ads_overall_summary x " + totalDataConString);

        //自定义回显对象 结果自动封装到给定的Entity对象中
        totalDataSql.setCallback(Sqls.callback.entity());
        totalDataSql.setEntity(dao.getEntity(AdsOverallSummaryVO.class));
        //执行自定义sql语句
        dao.execute(totalDataSql);
        //得到一个包含有总数据的AdsOverallSummaryVO对象
        AdsOverallSummaryVO totalDataVO = totalDataSql.getObject(AdsOverallSummaryVO.class);

        //获取查询条件 查询今日数据 需要按当前日期查询
        Condition todayDataCon = myCondition(param, false, 1);
        //查询条件转换为sql语句 设置映射 Entity为 AdsOverallSummary对象
        String todayDataConString = todayDataCon.toSql(dao.getEntity(AdsOverallSummary.class));
        Sql todayDataSql = Sqls.create("SELECT  \n" +
                "IFNULL(SUM(x.today_player_count),0) today_player_count ,\n" +
                "IFNULL(SUM(x.yesterday_player_count),0) yesterday_player_count ,\n" +
                "IFNULL(SUM(x.today_total_cost),0) today_total_cost ,\n" +
                "IFNULL(SUM(x.yesterday_total_cost),0) yesterday_total_cost ,\n" +
                "IFNULL(SUM(x.today_new_player_amount),0) today_new_player_amount ,\n" +
                "IFNULL(SUM(x.today_amount),0) today_amount ,\n" +
                "IFNULL(SUM(x.yesterday_new_player_amount),0) yesterday_new_player_amount ,\n" +
                "IFNULL(SUM(x.yesterday_amount),0) yesterday_amount ,\n" +
                "IFNULL(SUM(x.today_agent_count),0) today_agent_count ,\n" +
                "IFNULL(SUM(x.yesterday_agent_count),0) yesterday_agent_count ,\n" +
                "IFNULL(SUM(x.today_game_count),0) today_game_count ,\n" +
                "IFNULL(SUM(x.yesterday_game_count),0) yesterday_game_count ,\n" +
                "IFNULL(SUM(x.first_new_user_amount)/SUM(x.today_total_cost),0) first_roi \n" +
                "FROM ads_overall_summary x" + todayDataConString);

        //自定义回显对象 结果自动封装到给定的Entity对象中
        todayDataSql.setCallback(Sqls.callback.entity());
        todayDataSql.setEntity(dao.getEntity(AdsOverallSummaryVO.class));
        //执行自定义sql语句
        dao.execute(todayDataSql);
        //得到一个包含有今日和昨日数据的AdsOverallSummaryVO对象
        AdsOverallSummaryVO adsOverallSummaryVO = todayDataSql.getObject(AdsOverallSummaryVO.class);

        //拼接参数
        copyNullProperties(totalDataVO, adsOverallSummaryVO);

        //7日和30日Roi数据需要手动计算
        //7日内新用户累计充值
        BigDecimal d7NewPlayerAmount = BigDecimal.valueOf(dao.func(AdsOverallSummary.class, "sum", "todayNewPlayerAmount", myCondition(param, false, 7)));
        //7日内总消耗
        BigDecimal d7TotalCost = BigDecimal.valueOf(dao.func(AdsOverallSummary.class, "sum", "todayTotalCost", myCondition(param, false, 7)));
        //计算7日Roi
        adsOverallSummaryVO.setD7TotalRoi(BigDecimal.ZERO.equals(d7TotalCost) ? BigDecimal.ZERO : d7NewPlayerAmount.divide(d7TotalCost, 4, RoundingMode.HALF_UP));

        //30日内新用户累计充值
        BigDecimal d30NewPlayerAmount = BigDecimal.valueOf(dao.func(AdsOverallSummary.class, "sum", "todayNewPlayerAmount", myCondition(param, false, 30)));
        //30日内总消耗
        BigDecimal d30TotalCost = BigDecimal.valueOf(dao.func(AdsOverallSummary.class, "sum", "todayTotalCost", myCondition(param, false, 30)));
        //计算30日Roi
        adsOverallSummaryVO.setD30TotalRoi(BigDecimal.ZERO.equals(d30TotalCost) ? BigDecimal.ZERO : d30NewPlayerAmount.divide(d30TotalCost, 4, RoundingMode.HALF_UP));

        return adsOverallSummaryVO;
    }

    @Override
    public List<AdsOverallSummaryLineDataVO> getOverallSummaryLineData(AdsOverallSummaryParam param) {

        //根据条件获取数据
        Condition lineCon = myCondition(param, true, -1);
        //查询条件变为字符串
        String lineConString = lineCon.toSql(dao.getEntity(AdsOverallSummary.class));

        //自定义sql语句 加上条件字符串
        Sql sql = Sqls.create("Select dt," +
                "SUM(x.today_total_cost) cost, " +
                "SUM(x.today_amount) amount, " +
                "SUM(x.today_buy_amount) buy_amount," +
                "SUM(x.today_nature_amount) nature_amount, " +
                "IFNULL(SUM(x.first_new_user_amount)/SUM(x.today_total_cost),0) first_roi," +
                "SUM(x.today_player_count) player_account " +
                "from ads_overall_summary x " + lineConString);

        //自定义回显对象 结果自动添加到一个List中 List<Entity>
        sql.setCallback(Sqls.callback.entities());
        sql.setEntity(dao.getEntity(AdsOverallSummaryLineDataVO.class));

        //执行自定义sql语句
        dao.execute(sql);

        //判断查询日期的数据是否都存在 结果返回
        return checkAllDateData(sql.getList(AdsOverallSummaryLineDataVO.class), param);
    }

    /**
     * 自定义查询条件
     *
     * @param param         前端传递的查询参数实体
     * @param needLocalDate 是否拼接日期查询条件
     * @param queryDays     拼接需要查询数据的天数条件（给今日昨日、首日Roi、7日Roi以及30日Roi使用）
     * @return Condition 查询条件
     */
    private Condition myCondition(AdsOverallSummaryParam param, Boolean needLocalDate, Integer queryDays) {

        Criteria cri = Cnd.cri();
        //默认查询条件
        if (param.getGameId() != null) {
            //拼接游戏查询条件
            cri.where().andEquals("gameId", param.getGameId());
        }
        if (param.getPithcerId() != null) {
            //拼接投手查询条件
            cri.where().andEquals("pithcerId", param.getPithcerId());
        }
        if (param.getAgentId() != null) {
            //拼接渠道查询条件
            cri.where().andEquals("agentId", param.getAgentId());
        }
        if (param.getAccountId() != null) {
            //拼接广告账户查询条件
            cri.where().andEquals("accountId", param.getAccountId());
        }

        /*
         * 如果是组长 则需要排除掉 pitcher = -2 的所有数据
         * cri.where().andNotEquals("pithcerId", -2);
         * 如果是管理员账号 则不需要排除 pitcher = -2 的数据
         * */

        //根据参数拼接条件
        switch (queryDays) {
            case 1:
                cri.where().andEquals("dt", LocalDate.now());
                break;

            case 7:
                cri.where().andBetween("dt", LocalDate.now().minusDays(7), LocalDate.now());
                break;

            case 30:
                cri.where().andBetween("dt", LocalDate.now().minusDays(30), LocalDate.now());
                break;

            default:
                break;
        }

        //折线数据需要拼接时间查询条件
        if (needLocalDate) {
            //如果传入时间区间参数为空 设置默认值
            if (param.getBeginDate() == null || param.getEndDate() == null) {
                param.setBeginDate(LocalDate.now().minusDays(30));
                param.setEndDate(LocalDate.now());
            }
            //添加日期条件
            cri.where().andBetween("dt", param.getBeginDate(), param.getEndDate());
            //按日期分组
            cri.getGroupBy().groupBy("dt");
            //按日期排序 升序
            cri.getOrderBy().asc("dt");
        }

        return cri;
    }

    /**
     * 所有为空值的属性都不copy
     *
     * @param source 原数据
     * @param target 目标数据
     */
    private void copyNullProperties(Object source, Object target) {
        BeanUtils.copyProperties(source, target, getNullField(source));
    }

    /**
     * 获取属性中为空的字段
     *
     * @param target 目标对象
     * @return 不需要替换的字段数组
     */
    private static String[] getNullField(Object target) {
        BeanWrapper beanWrapper = new BeanWrapperImpl(target);
        PropertyDescriptor[] propertyDescriptors = beanWrapper.getPropertyDescriptors();
        Set<String> notNullFieldSet = new HashSet<>();
        if (propertyDescriptors.length > 0) {
            for (PropertyDescriptor p : propertyDescriptors) {
                String name = p.getName();
                Object value = beanWrapper.getPropertyValue(name);
                if (Objects.isNull(value)) {
                    notNullFieldSet.add(name);
                }
            }
        }
        String[] notNullField = new String[notNullFieldSet.size()];
        return notNullFieldSet.toArray(notNullField);
    }

    /**
     * 检查前端查询日期内的数据是否在数据库中都存在
     * 不存在要设置默认值 0
     *
     * @param list 需要判断的list
     * @param param 前端传递的查询参数实体
     * @return 返回给前端的折线图数据 list
     */
    private List<AdsOverallSummaryLineDataVO> checkAllDateData(List<AdsOverallSummaryLineDataVO> list, AdsOverallSummaryParam param) {

        //先判断是否缺少数据 不少数据直接返回
        if ((list.size() - 1) == param.getBeginDate().until(param.getEndDate(), ChronoUnit.DAYS)) {
            return list;
        }

        //创建一个Map用来存储日期数据
        Map<LocalDate, Integer> map = new HashMap<>();
        //添加日期数据
        LocalDate tempDate = param.getBeginDate();
        while (!tempDate.isAfter(param.getEndDate())) {
            map.put(tempDate, 1);
            tempDate = tempDate.plusDays(1);
        }
        //循环遍历list 查询vo里的日期参数
        for (AdsOverallSummaryLineDataVO vo : list) {
            Integer value = map.get(vo.getDt());
            if (value != null) {
                map.put(vo.getDt(), ++value);
            }
        }
        //遍历map数据 value为 1 的表示 list中没有该日期的实体
        for (Map.Entry<LocalDate, Integer> entry : map.entrySet()) {
            if (entry.getValue() == 1) {
                //list数据添加一个默认VO对象
                list.add(AdsOverallSummaryLineDataVO.builder()
                        .dt(entry.getKey())
                        .amount(BigDecimal.ZERO)
                        .buyAmount(BigDecimal.ZERO)
                        .natureAmount(BigDecimal.ZERO)
                        .cost(BigDecimal.ZERO)
                        .firstRoi(BigDecimal.ZERO)
                        .playerAccount(BigInteger.ZERO)
                        .build());
            }
        }
        //排序 lambda表达式 按日期升序排序
        list.sort((vo1, vo2) -> vo1.getDt().isBefore(vo2.getDt()) ? -1 : 1);

        return list;
    }


}
